/* 
 * Stellarium
 * Copyright (C) 2002 Fabien Chreau
 * Inspired by the gui.h by Chris Laurel <claurel@shatters.net>
 * in his Open Source Software Celestia
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "s_gui.h"

#include <stdio.h>

using namespace std;
using namespace gui;


/**** GraphicsContext ****/

GraphicsContext::GraphicsContext(int _winW, int _winH) :
	winW(_winW),
	winH(_winH),
	font(NULL), 
	backGroundTexture(1), 
	headerTexture(1), 
	baseColor(vec3_t(0.2,0.2,0.4)),
	textColor(vec3_t(0.8,0.8,1.0)),
	scissorPos(vec2_t(0,0))
{
}

s_font* GraphicsContext::getFont() const
{
    return font;
}

void GraphicsContext::setFont(s_font* _font)
{
    font = _font;
}


/**** Component ****/

Component::Component() :
    position(0, 0),
    size(0, 0),
    parent(NULL),
	clicCallback(NULL),
	moveCallback(NULL),
	mouseIn(false),
	visible(true),
	draging(false),
	active(true),
	ID(-1),
	passThru(false)
{
}

vec2_t Component::getPosition() const
{	return position;
}

vec2_t Component::getSize() const
{	return size;
}

void Component::reshape(vec2_t _position, vec2_t _size)
{	position = _position;
	size = _size;
}

Component* Component::getParent() const
{	return parent;
}

void Component::setParent(Component* c)
{	parent = c;
}

void Component::setClicCallback(void (*_clicCallback)(int,int,enum guiValue,enum guiValue, Component *))
{	clicCallback=_clicCallback;
}

void Component::setMoveCallback(void (*_moveCallback)(int,int,enum guiValue, Component *))
{	moveCallback=_moveCallback;
}

/**** Container ****/

void containerClicCallBack(int x,int y,enum guiValue state,enum guiValue button,Component * caller)
{	((Container *)caller)->handleMouseClic((int)(x-caller->getPosition()[0]),(int)(y-caller->getPosition()[1]),state,button);
}

void containerMoveCallBack(int x,int y,enum guiValue,Component * caller)
{	((Container *)caller)->handleMouseMove((int)(x-caller->getPosition()[0]),(int)(y-caller->getPosition()[1]));
}

Container::Container() : Component()
{	setClicCallback(containerClicCallBack);
	setMoveCallback(containerMoveCallBack);
}

Container::~Container()
{	vector<Component*>::iterator iter = components.begin();
    while (iter != components.end())
    {	delete (*iter);
		iter++;
	}
}

int Container::getComponentCount() const
{
    return components.size();
}

Component* Container::getComponent(int n) const
{
    if (n >= 0 && n < components.size())
        return components[n];
    else
        return NULL;
}

void Container::addComponent(Component* c)
{	// ASSERT(c->getParent() == NULL)
	components.insert(components.end(), c);
	c->setParent(this);
}

void Container::render(GraphicsContext& gc)
{	if (!visible) return;
	vector<Component*>::iterator iter = components.begin();
	glPushMatrix();
	glTranslatef(getPosition()[0], getPosition()[1], 0);
	vec2_t pos = getPosition();
	vec2_t sz = getSize();
	gc.scissorPos+=pos;
    while (iter != components.end())
    {	if ((*iter)->visible)
		{	glScissor((int)gc.scissorPos[0],(int)(gc.winH-gc.scissorPos[1]-sz[1]),(int)sz[0],(int)sz[1]);
			(*iter)->render(gc);
		}
		iter++;
    }
	gc.scissorPos-=pos;
    glPopMatrix();
}

int Container::handleMouseClic(int x, int y, enum guiValue state, enum guiValue button)
{	vector<Component*>::iterator iter = components.begin();
    while (iter != components.end())
    {	if ((*iter)->visible && (*iter)->clicCallback!=NULL && ((*iter)->draging || (*iter)->isIn(x,y)))
		{	(*iter)->clicCallback(x,y,state,button,(*iter));
			if (!(*iter)->passThru) return 1;
		}
		iter++;
	}
	return 0;
}

void Container::handleMouseMove(int x, int y)
{	vector<Component*>::iterator iter = components.begin();
    while (iter != components.end())
    {
		if ((*iter)->visible && (*iter)->moveCallback!=NULL)
		{	if((*iter)->draging || (*iter)->isIn(x,y))
			{	if (!(*iter)->mouseIn)
				{	(*iter)->moveCallback(x,y,GUI_MOUSE_ENTER,(*iter));
					(*iter)->mouseIn=true;
				}
				else
				{	(*iter)->moveCallback(x,y,GUI_NOTHING,(*iter));
				}
			}
			else
			{	if ((*iter)->mouseIn) (*iter)->moveCallback(x,y,GUI_MOUSE_LEAVE,(*iter));
				(*iter)->mouseIn=false;
			}
		}
		iter++;
	}
}

/*** FilledContainer ***/

void FilledContainer::render(GraphicsContext& gc)
{	glColor3fv(gc.baseColor);
	vec2_t pos = getPosition();
	vec2_t sz = getSize();
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBindTexture(GL_TEXTURE_2D,gc.backGroundTexture);
	glBegin(GL_QUADS );
		glTexCoord2f(0.0f,0.0f);	glVertex3f(pos[0],			pos[1] + sz[1],	0);	// Bas Gauche
		glTexCoord2f(1.0f,0.0f);	glVertex3f(pos[0] + sz[0],	pos[1] + sz[1],	0);	// Bas Droite
		glTexCoord2f(1.0f,1.0f);	glVertex3f(pos[0] + sz[0],	pos[1],			0);	// Haut Droit
		glTexCoord2f(0.0f,1.0f);	glVertex3f(pos[0],			pos[1],			0);	// Haut Gauche
	glEnd ();
	Container::render(gc);
}


/**** Button ****/

void ExtButtonClicCallback(int x,int y,enum guiValue state,enum guiValue button,Component * me)
{	((Button*)me)->ButtonClicCallback(button, state);
}

void ExtButtonMoveCallback(int x,int y,enum guiValue action,Component * me)
{	((Button*)me)->ButtonMoveCallback(action);
}

Button::Button() : Component(), mouseOn(false), onClicCallback(NULL), onMouseOverCallback(NULL)
{	setClicCallback(ExtButtonClicCallback);
	setMoveCallback(ExtButtonMoveCallback);
}

void Button::render(GraphicsContext& gc)
{	if (mouseOn) glColor3fv(gc.baseColor/2);
	else glColor3fv(gc.baseColor);
	glDisable(GL_TEXTURE_2D);
	glBegin(GL_LINE_LOOP);
	vec2_t pos = getPosition();
	vec2_t sz = getSize();
	glVertex3f(pos[0],        pos[1],        0);
	glVertex3f(pos[0] + sz[0], pos[1],        0);
	glVertex3f(pos[0] + sz[0], pos[1] + sz[1], 0);
	glVertex3f(pos[0],        pos[1] + sz[1], 0);
	glEnd();
}

void Button::setOnClicCallback(void (*_onClicCallback)(enum guiValue,Component *))
{	onClicCallback=_onClicCallback;
}

void Button::setOnMouseOverCallback(void (*_onMouseOverCallback)(enum guiValue,Component *))
{	onMouseOverCallback=_onMouseOverCallback;
}

void Button::ButtonClicCallback(enum guiValue button,enum guiValue state)
{	if (state==GUI_DOWN) mouseOn=true;
	else
	{	if (mouseOn && onClicCallback!=NULL) onClicCallback(button,this);
		mouseOn=false;
	}
}

void Button::ButtonMoveCallback(enum guiValue action)
{	if (action==GUI_MOUSE_LEAVE || action==GUI_MOUSE_ENTER)
	{	mouseOn=false;
	}
	if (onMouseOverCallback!=NULL) onMouseOverCallback(action,this);
}


/**** Labeled_Button ****/

Labeled_Button::Labeled_Button(char * _label) : Button()
{	setLabel(_label);
}

Labeled_Button::~Labeled_Button()
{	delete label;
}

const char * Labeled_Button::getLabel() const
{	return label;
}

void Labeled_Button::setLabel(char * _label)
{	if (label!=NULL) delete label;
	label=strdup(_label);
}

void Labeled_Button::render(GraphicsContext& gc)
{	Button::render(gc);
	vec2_t pos = getPosition();
	vec2_t sz = getSize();
	if (mouseOn) glColor3fv(gc.textColor/2);
	else glColor3fv(gc.textColor);
    glEnable(GL_TEXTURE_2D);
    glPushMatrix();
    glTranslatef(pos[0] + (sz[0] - gc.getFont()->getStrLen(label)) / 2,
                 pos[1] + (sz[1] - gc.getFont()->getLineHeight()) / 2,
                 0);
    gc.getFont()->print(0,0,label);
    glPopMatrix();
}

/**** Textured_Button ****/

Textured_Button::Textured_Button(unsigned int _texID)
: Button(), texID(_texID), activeColor(1.0,1.0,1.0), passiveColor(0.5,0.5,0.5)
{
}

Textured_Button::Textured_Button(unsigned int _texID, vec2_t _position, vec2_t _size 
,vec3_t _activeColor, vec3_t _passiveColor, 
void (*_onClicCallback)(enum guiValue,Component *),
void (*_onMouseOverCallback)(enum guiValue,Component *), int _ID, bool _active)
	: Button(), texID(_texID),
	activeColor(_activeColor), passiveColor(_passiveColor)
{	position=_position;
	size=_size;
	setOnClicCallback(_onClicCallback);
	setOnMouseOverCallback(_onMouseOverCallback);
	ID=_ID;
	active=_active;
}

void Textured_Button::render(GraphicsContext& gc)
{	if (active)
	{	glColor3fv(activeColor);
	}
	else 
	{	glColor3fv(passiveColor);
	}
	vec2_t pos = getPosition();
	vec2_t sz = getSize();
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBindTexture(GL_TEXTURE_2D,texID);
	glBegin(GL_QUADS );
		glTexCoord2f(0.0f,0.0f);	glVertex3f(pos[0],			pos[1] + sz[1],	0);	// Bas Gauche
		glTexCoord2f(1.0f,0.0f);	glVertex3f(pos[0] + sz[0],	pos[1] + sz[1],	0);	// Bas Droite
		glTexCoord2f(1.0f,1.0f);	glVertex3f(pos[0] + sz[0],	pos[1],			0);	// Haut Droit
		glTexCoord2f(0.0f,1.0f);	glVertex3f(pos[0],			pos[1],			0);	// Haut Gauche
	glEnd ();
}

/**** Label ****/

Label::Label(char * _label) : Component(), label(NULL), theFont(NULL),colour(vec3_t(-1,-1,-1))
{	setLabel(_label);
	passThru=true;
}

Label::Label(char * _label, s_font * _theFont) : Component(), label(NULL), theFont(_theFont), colour(vec3_t(-1,-1,-1))
{	setLabel(_label);
	passThru=true;
}

Label::~Label()
{	delete label;
}

const char * Label::getLabel() const
{	return label;
}

void Label::setLabel(char * _label)
{	if (label!=NULL) delete label;
	label=strdup(_label);
}

void Label::render(GraphicsContext& gc)
{	glEnable(GL_TEXTURE_2D);
	if (colour[0]<0) glColor3fv(gc.textColor);
	else 
	glColor3fv(colour);
	vec2_t pos = getPosition();
	vec2_t sz = getSize();
	glPushMatrix();
	if (theFont==NULL)
	{	glTranslatef(pos[0],pos[1] + (sz[1] - gc.getFont()->getLineHeight()) / 2,0);
		gc.getFont()->print(0,0,label);
	}
	else
	{	glTranslatef(pos[0],pos[1] + (sz[1] - theFont->getLineHeight()) / 2,0);
		theFont->print(0,0,label);
	}
	glPopMatrix();
}

/**** TextLabel ****/

TextLabel::TextLabel(char * _label, s_font * _theFont) : Container(), label(NULL), theFont(_theFont), colour(vec3_t(-1,-1,-1))
{	passThru=true;
	if (theFont==NULL) return;
	setLabel(_label);
}

TextLabel::~TextLabel()
{	if (label!=NULL) delete label;
//	Container::~Container();
}

const char * TextLabel::getLabel() const
{	return label;
}

void TextLabel::setLabel(char * _label)
{	if (label!=NULL) delete label;
	if (_label==NULL) return;
	label=strdup(_label);
	components.clear();
	Label * tempLabel;
	char * pch;
	int i = 0;
	pch = strtok (label,"\n");
	while (pch != NULL)
	{	tempLabel = new Label(pch,theFont);
		tempLabel->reshape(vec2_t(0,i*(theFont->getLineHeight()+1)+3),vec2_t(theFont->getStrLen(pch),theFont->getLineHeight()));
		tempLabel->setColour(colour);
		addComponent(tempLabel);
		pch = strtok (NULL,"\n");
		i++;
	}
}

void TextLabel::setColour(vec3_t _colour)
{	colour=_colour;
	vector<Component*>::iterator iter = components.begin();
	while (iter != components.end())
	{	{	((Label*)(*iter))->setColour(_colour);
		}
		iter++;
	}
}

/**** FilledTextLabel ****/

FilledTextLabel::FilledTextLabel(char * _label, s_font * _theFont) : TextLabel(_label,_theFont)
{
}

FilledTextLabel::~FilledTextLabel()
{	if (label!=NULL) delete label;
}

void FilledTextLabel::render(GraphicsContext& gc)
{	glColor3fv(gc.baseColor);
	vec2_t pos = getPosition();
	vec2_t sz = getSize();
	glPushMatrix();
	glDisable(GL_TEXTURE_2D);
	glBegin(GL_LINE_LOOP);
	glVertex3f(pos[0],        pos[1],        0);
	glVertex3f(pos[0] + sz[0], pos[1],        0);
	glVertex3f(pos[0] + sz[0], pos[1] + sz[1], 0);
	glVertex3f(pos[0],        pos[1] + sz[1], 0);
	glEnd();
	glPopMatrix();

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBindTexture(GL_TEXTURE_2D,gc.backGroundTexture);
	glBegin(GL_QUADS );
		glTexCoord2f(0.0f,0.0f);	glVertex3f(pos[0],			pos[1] + sz[1],	0);	// Bas Gauche
		glTexCoord2f(1.0f,0.0f);	glVertex3f(pos[0] + sz[0],	pos[1] + sz[1],	0);	// Bas Droite
		glTexCoord2f(1.0f,1.0f);	glVertex3f(pos[0] + sz[0],	pos[1],			0);	// Haut Droit
		glTexCoord2f(0.0f,1.0f);	glVertex3f(pos[0],			pos[1],			0);	// Haut Gauche
	glEnd ();
	TextLabel::render(gc);
}
